import logging
from abc import ABC, abstractmethod
from collections import OrderedDict
from pathlib import Path

from devman.common import assert_path_exist_and_is_dir
from devman.file.iter import iter_dirs, iter_files
from devman.helper.table import build_table

log = logging.getLogger(__name__)


class BaseStat(ABC):
    def __init__(self, max_depth: int = 16):
        self._max_depth = max_depth
        self.res = {}

    @abstractmethod
    def handle_path(self, path: Path) -> bool: ...

    # 遍历
    def stat_paths(self, root: Path):
        for file in iter_files(root, self._max_depth):
            if self.handle_path(file) is False:
                break

    def build_table(self): ...


class SuffixStat(BaseStat):
    def __init__(self, suffix_set: set[str], max_cnt: int = 4096, max_depth: int = 16):
        super().__init__(max_depth=max_depth)
        self._max_cnt = max_cnt
        self._cur_cnt = 0
        self._suffix_set = suffix_set
        self._res = {}

    def _match_suffix(self, suffix: str):
        if self._suffix_set is None:
            return True
        return suffix in self._suffix_set

    def handle_path(self, path):
        self._cur_cnt += 1
        if self._cur_cnt > self._max_cnt:
            return False
        _suffix = path.suffix
        if path.is_file() and self._match_suffix(_suffix):
            cur = self._res.setdefault(_suffix, 0)
            self._res[_suffix] = cur + 1
        return True

    def build_table(self):
        header = ["文件类型", "数量"]
        rows = [(k, str(v)) for k, v in self._res.items()]
        rows.append(("TOTAL", str(self._cur_cnt)))
        table = build_table("根据 SUFFIX 计数", header, rows)
        return table


def api_stat_prefix(root: Path, ext: list[str], max_depth: int = 16):
    res = {}
    ext_set = set(e.lower() for e in ext)

    for file in iter_files(root, max_depth):
        file: Path
        if file.suffix.lower() in ext_set:
            prefix = file.stem.split("_")[0]
            val = res.get(prefix, 0)
            res[prefix] = val + 1

    header = ["前缀 PREFIX", "数目"]
    rows = [(k, str(v)) for k, v in res.items()]
    table = build_table("根据目录统计文件", header, rows)
    log.info(table)


def api_stat_info_in_dir(path: Path) -> tuple[int, int]:
    assert path.is_dir()
    file_cnt = dir_cnt = other_cnt = 0
    for item in path.iterdir():
        if item.is_file():
            file_cnt += 1
        elif item.is_dir():
            dir_cnt += 1
        else:
            other_cnt += 1
    return file_cnt, dir_cnt, other_cnt


def api_stat_cnt(root: Path, max_depth: int = 16):
    res: dict[Path, tuple] = OrderedDict()

    def adder(path: Path, f, d, o):
        parts = path.relative_to(root).parts
        for i in range(len(parts)):
            _path = root.joinpath(*parts[:i])
            _f, _d, _o = res.get(_path, (0, 0, 0))
            res[_path] = (f + _f, d + _d, o + _o)

    for dir in iter_dirs(root, max_depth=max_depth):
        f, d, o = api_stat_info_in_dir(dir)
        res[dir] = (f, d, o)
        adder(dir, f, d, o)
    header = ["路径", "文件数", "目录数", "其他文件"]
    rows = [
        (str(k.relative_to(root)), str(f), str(d), str(o))
        for k, (f, d, o) in res.items()
    ]
    table = build_table("根据目录统计文件", header, rows)
    log.info(table)


def stat(src: Path, max_depth: int = 4):
    """统计文件夹个数大小，文件个数大小"""
    assert_path_exist_and_is_dir(src)
    # 用队列解决
    total_file = 0
    total_folder = 0
    queue = [(src, 0)]
    while queue:
        p, level = queue.pop()
        if level > max_depth:  # 如果递归深度大于最大深度，跳出递归
            continue
        try:
            for p in p.iterdir():
                if p.is_dir():
                    total_folder += 1
                    queue.append((p, level + 1))  # 递归进入下一层
                else:
                    total_file += 1
        except PermissionError:
            log.info(f"无权限访问，跳过... {p}")
            continue
        except Exception as e:
            log.error("未知错误...")
            log.exception(e)
    # print(f"文件夹 : {src} ")
    # print(f"文件数量   : {total_file}")
    # print(f"文件夹数量 : {total_folder}")
    return src, total_file, total_folder


class CounterStat(BaseStat):
    def __init__(self, suffix_set: set[str], max_cnt: int = 4096, max_depth: int = 16):
        super().__init__(max_depth=max_depth)
        self._max_cnt = max_cnt
        self._cur_cnt = 0
        self._suffix_set = suffix_set
        self._res = {}

    def handle_path(self, path):
        self._cur_cnt += 1
        if self._cur_cnt > self._max_cnt:
            return False
        _suffix = path.suffix
        if path.is_file():
            cur = self._res.setdefault(_suffix, 0)
            self._res[_suffix] = cur + 1
        return True

    def build_table(self):
        header = ["文件类型", "数量"]
        rows = [(k, str(v)) for k, v in self._res.items()]
        rows.append(("TOTAL", str(self._cur_cnt)))
        table = build_table("根据 SUFFIX 计数", header, rows)
        return table



from pathlib import Path
from typing import Callable

import inquirer

from devman.helper.cache import JsonCache

cache = JsonCache(Path(".cache", "cache.json"))


def query_list(func: str, name: str, msg: str, validate: Callable = None):
    choices = cache.arg_get_list(func, name)
    if choices:
        res = inquirer.list_input(msg, choices=choices, other=True)
    else:
        res = inquirer.text(msg)
    # if validate and validate(res):
    if validate is None or validate(res):
        cache.arg_upsert(func, name, res)
    return res


def query_check(func: str, name: str, msg: str):
    choices = cache.arg_get_list(func, name)
    res = inquirer.checkbox(msg, choices=choices, other=True)
    cache.arg_upsert_list(func, name, res)
    return res



"""
create_time: 2025-06-22 16:12:08
author: jackylee
"""

from abc import ABC, abstractmethod
from collections import OrderedDict

# from cachetools import LRUCache
# TODO 了解其功能


class LRU(ABC):
    @abstractmethod
    def get(self, key: str) -> object: ...

    @abstractmethod
    def put(self, key: str, val: object) -> bool: ...


class BuildInLRU(LRU):
    """使用 python 内置 OrderedDict 实现 LRU"""

    def __init__(self, capacity=10):
        self.capacity = capacity
        self.cache = OrderedDict()

    def get(self, key):
        if key not in self.cache.keys():
            return None
        self.cache.move_to_end(key)
        return self.cache[key]

    def put(self, key, val):
        if key in self.cache.keys():
            self.cache.move_to_end(key)
        self.cache[key] = val
        if len(self.cache) > self.capacity:
            self.cache.popitem(False)



"""
Author: Jacky Lee
Date: 2025-10-22
Description: 递归解析 json 代码
"""

import json
import logging
from typing import Union

log = logging.getLogger(__name__)


def recursive_parse(
    arg: Union[dict, list, str, int, float, None],
) -> Union[dict, list, str, int, float, None]:
    """递归解析,参数为 dict, list, str, int, float 中的一种

    Args:
        param (Union[dict, list, str, int, float]):

    Raises:
        TypeError: 类型错误

    Returns:
        Union[dict, list, str, int, float]: 返回类型是该五种基本类型
    """
    if arg is None:
        return arg
    if isinstance(arg, list):
        return parse_list(arg)
    elif isinstance(arg, dict):
        return parse_dict(arg)
    elif isinstance(arg, str):
        return parse_str(arg)
    elif isinstance(arg, int):
        return arg
    elif isinstance(arg, float):
        return arg
    else:
        raise TypeError("Argument is not any of List, Object, String, Number or None.")


def parse_str(text: str) -> Union[dict, list, str]:
    try:
        # 如果可以被解析
        shallow_parsed = json.loads(text)
    except json.JSONDecodeError:
        return text
    deep_parsed = recursive_parse(shallow_parsed)
    return deep_parsed


def parse_list(li: list) -> list:
    _list = []
    for _, item in enumerate(li):
        tmp = recursive_parse(item)
        _list.append(tmp)
    return _list


def parse_dict(di: dict) -> dict:
    _di = {}
    for k, v in di.items():
        _di[k] = recursive_parse(v)
    return _di



import logging
from pathlib import Path

import pyperclip
from rich.console import Console

from devman.common import assert_path_exist_and_is_file

log = logging.getLogger(__name__)
console = Console()


def confirm() -> bool:
    while True:
        flag = input("是否继续(y/n)...").strip().lower()  # 获取用户输入并处理为小写
        if flag.startswith("y"):  # 如果以'y'开头
            log.info("继续执行...")
            return True
        elif flag.startswith("n"):  # 如果以'n'开头
            log.info("退出程序...")
            return False
        log.info("无效输入，请输入'y'或'n'。")  # 提示用户输入无效，继续询问


def from_clipboard_or_file(src: Path) -> str:
    if src is None:
        return pyperclip.paste()
    assert_path_exist_and_is_file(src)
    return src.read_text()


def to_clipboard_or_file(dst: Path, content: str, force: bool, quiet=False) -> bool:
    # 写入剪贴板
    console.quiet = quiet
    if dst is None:
        pyperclip.copy(content)
        console.print("已复制到剪贴板")
        return
    # dst 非空,路径
    if dst.exists() and force:
        dst.write_text(content)
        console.print(f"已写入文件 {dst.name}")
    else:
        console.print(f"目标文件 {dst.name} 已存在，使用 -f 强制输出")



import logging

import typer
from rich.console import Console
from typing_extensions import Annotated

from devman.args import ARG_DST, ARG_FORCE_COVER_DST, ARG_QUIET, ARG_SRC, ARG_VERBOSE
from devman.helper.interactive import from_clipboard_or_file, to_clipboard_or_file
from devman.json.api import api_json_dump_obj_to_str, api_parse_str_to_json
from devman.log import config_log

app = typer.Typer()
console = Console()

ARG_RECURSIVE = Annotated[
    bool, typer.Option(help="是否递归去转义", show_default="默认递归")
]


@app.command("parse", help="解析字符串为 json(默认递归去转义)")
def recursive_parse_json(
    src: ARG_SRC = None,
    dst: ARG_DST = None,
    recursive: ARG_RECURSIVE = True,
    force: ARG_FORCE_COVER_DST = False,
    verbose: ARG_VERBOSE = False,
    quiet: ARG_QUIET = False,
):
    console.quiet = quiet
    if verbose:
        console.rule("解析 json 字符串")
        console.print("开启详细输出")
        config_log(logging.DEBUG)
    dump_content = None
    try:
        origin_content = from_clipboard_or_file(src)
        parsed_content = api_parse_str_to_json(origin_content, recursive)
        dump_content = api_json_dump_obj_to_str(parsed_content)
        to_clipboard_or_file(dst, dump_content, force, quiet)
    except AssertionError as e:
        console.print("断言错误", e)
    except Exception as e:
        console.print("未知异常", e)
        console.print("使用 -v 详细输出")
    console.print(dump_content)


if __name__ == "__main__":
    app()



import json
import re
from datetime import datetime
from typing import Union

from devman.json.core import parse_str, recursive_parse


def api_json_dump_obj_to_str(obj: Union[dict, list]):
    return json.dumps(obj, indent=2, ensure_ascii=False)


def api_parse_str_to_json(
    arg: str, recursive: bool
) -> Union[dict, list, str, int, float]:
    # 递归解析
    if recursive:
        return recursive_parse(arg)
    # 正常解析
    return parse_str(arg)


def get_possible_datetime_from_str(line: str) -> datetime:
    raise NotImplementedError("Not Finish.")
    dt_pat = re.compile(r"(\d{4}-?\d{2}-?\d{2}\s?\d{2}:?\d{2}:?\d{2}(\.\d{3})?)")
    return dt_pat


def get_possible_json_from_str(line: str) -> dict:
    raise NotImplementedError("Not Finish.")
    json_pat = re.compile(r"(?<!#)(\[?{.*}\]?)(?!#)")
    return json_pat


def parse_lines(lines: list[str]) -> list:
    raise NotImplementedError("Not Finish.")
    res = []
    for idx, line in enumerate(lines):
        get_possible_datetime_from_str(line)
        get_possible_json_from_str(line)
    return res



from pathlib import Path
from typing import Generator

from devman.common import assert_path_exist_and_is_dir, is_dot_path


def iter_dirs(path: Path, max_depth: int):
    _path = path.absolute()
    assert_path_exist_and_is_dir(_path)
    q = [(_path, 0)]
    while q:
        _dir, depth = q.pop(0)
        if depth > max_depth:
            continue
        for item in _dir.iterdir():
            if item.is_dir():
                q.append((item, depth + 1))
        yield _dir


def iter_files(path: Path, max_depth: int) -> Generator[Path, None, None]:
    _path = path.absolute()
    assert_path_exist_and_is_dir(_path)
    q = [(_path, 0)]
    while q:
        _path, depth = q.pop(0)
        if depth > max_depth:
            continue
        if _path.is_file():
            yield _path
        elif _path.is_dir() and not is_dot_path(_path):
            for item in _path.iterdir():
                q.append((item, depth + 1))



from pathlib import Path

import typer
from typing_extensions import Annotated

ARG_SRC = Annotated[
    Path, typer.Argument(help="源文件路径", show_default="默认从剪贴板读取")
]
ARG_DST = Annotated[
    Path, typer.Argument(help="目标文件路径", show_default="默认输出到剪贴板")
]

ARG_FORCE_COVER_DST = Annotated[
    bool,
    typer.Option(
        "--force", "-f", help="是否强制覆盖 DST 目录", show_default="默认不强制"
    ),
]

ARG_VERBOSE = Annotated[
    bool, typer.Option("--verbose", "-v", help="详细输出", show_default=False)
]

ARG_QUIET = Annotated[
    bool, typer.Option("--quiet", "-q", help="静默输出", show_default=False)
]



from __future__ import annotations

from typing import Optional


class ListNode:
    def __init__(
        self, val: object = None, next_node: Optional[ListNode] = None
    ) -> None:
        self.val = val
        self.next_node = next_node


class LinkList:
    def __init__(self) -> None:
        """
        create linklist with head
        """
        # 哨兵节点
        self.head = self.tail = ListNode(None)

    def traverse_linklist(self, call=print) -> None:
        """
        traverse link list
        """
        node = self.head
        while node:
            call(node.val)
            node = node.next_node

    @staticmethod
    def build_from_list(nums: list[object]) -> None:
        """
        build linklist with list
        """
        ll = LinkList()
        if nums is None:
            return ll

        for idx, item in enumerate(nums):
            node = ListNode(item, None)
            if idx == 0:
                ll.head.next_node = node
            ll.tail = node



import logging
import re
from pathlib import Path

from devman.common import (
    assert_path_exist_and_is_dir,
    assert_path_exist_and_is_file,
    assert_path_not_exist_and_is_file,
)

log = logging.getLogger(__name__)


def _move_file(src: Path, dst: Path, dry: bool):
    assert_path_exist_and_is_file(src)
    assert_path_not_exist_and_is_file(dst)
    log.debug(f"(dry={dry})移动文件({not dry}: {src} -> {dst}")
    if not dry:
        src.rename(dst)


def move_match_pattern_file(
    src: Path,
    dst: Path = None,
    pattern: str = None,
    dry: bool = True,
    max_cnt: int = 16,
) -> bool:
    """移动指定的文件名"""
    # 校验
    assert_path_exist_and_is_dir(src)
    assert_path_exist_and_is_dir(dst)
    assert pattern is not None
    # 操作
    log.debug(f"提取文件到目录: {src}")
    pat = re.compile(pattern, re.IGNORECASE)
    total_cnt = 0
    for src_path in src.rglob("*"):
        total_cnt += 1
        if total_cnt > max_cnt:
            continue
        if src_path.is_file() and pat.search(src_path.stem):
            dst_path = dst.joinpath(src_path.name)
            _move_file(src_path, dst_path, dry)
    # 总结
    log.debug(f"总共移动文件({not dry}): {total_cnt} 个(最大移动文件为 {max_cnt})")
    return total_cnt


def move_prefix_ext(
    src: Path, dst: Path, *, prefix: re.Pattern, ext: list[str] = None, dry: bool = True
):
    ext_set = set([e.lower() for e in ext])
    cnt = exist = not_match = 0
    for file in src.rglob("*"):
        if not file.is_file():
            continue
        # log.debug(file, prefix.match(file.stem), file.suffix.lower() in ext_set)
        if file.suffix.lower() in ext_set and prefix.match(file.stem):
            dst_file = dst.joinpath(file.name)
            log.debug(f"移动({not dry}) {file} -> {dst_file}")
            if not dst_file.exists():
                cnt += 1
                if not dry:
                    file.rename(dst_file)
            else:
                exist += 1
                log.debug(f"{file} -> {dst_file} 文件已存在，无法移动")
        else:
            log.debug(f"文件={file} 不匹配")
            not_match += 1
    log.debug(
        f"总共移动文件数量={cnt}, 无法移动的文件数量={exist}, 不匹配文件数量={not_match}"
    )



import logging
import time
from os import PathLike
from pathlib import Path
from typing import Union

from tinydb import Query, TinyDB

# 数据库连接
log = logging.getLogger(__name__)


class JsonCache:
    """json 数据库, 使用 json 文件存储，方便直接修改"""

    KEY_NAME = "key"
    VAL_NAME = "val"
    TIMESTAMP = "timestamp"

    def __init__(self, path: PathLike):
        log.debug(f"连接数据库: {path}")
        _parent = Path(path).parent
        _parent.mkdir(parents=True, exist_ok=True)
        self._cache = TinyDB(path)
        self._query = Query()

    def set(self, key: str, val: Union[str, dict], tbl: str = None) -> None:
        """设置键值对

        Args:
            key (str): 键
            val (Union[str, dict]): 值
            tbl (str, optional): 表名, 默认为 None
        """
        _tbl = self._cache.table(tbl) if tbl else self._cache
        _key = self._query.__getattr__(self.KEY_NAME)
        _tbl.upsert({self.KEY_NAME: key, self.VAL_NAME: val}, _key == key)

    def get(self, key: str, tbl: str = None, default: Union[str, dict, list] = None):
        """获取键值对

        Args:
            key (str): 键
            default (Union[str, dict, list], optional): 默认值, 默认为 None.
            tbl (str, optional): 表名, 默认为 None.
        """
        _tbl = self._cache.table(tbl) if tbl else self._cache
        _key = self._query.__getattr__(self.KEY_NAME)
        _res = _tbl.get(_key == key)
        if _res:
            return _res.get(self.VAL_NAME)
        return default

    def get_all(self, tbl: str = None) -> dict:
        """在一个表中获取所有值
        Args:
            tbl (str, optional): 表名. 默认为 None.

        Returns:
            _type_: _description_
        """
        table = self._cache.table(tbl) if tbl else self._cache
        _res = {}
        for item in table.all():
            _key = item.get(self.KEY_NAME)
            _val = item.get(self.VAL_NAME)
            _res[_key] = _val
        return _res

    def arg_upsert(self, func: str, key: str, val):
        """参数插入"""
        table = self._cache.table(func)
        item = {self.KEY_NAME: key, self.VAL_NAME: val, self.TIMESTAMP: time.time()}
        q = (self._query.__getattr__(self.KEY_NAME) == key) & (
            self._query.__getattr__(self.VAL_NAME) == val
        )
        table.upsert(item, q)

    def arg_upsert_list(self, func: str, name: str, d: list):
        """参数批量插入 LIST"""
        for v in d:
            self.arg_upsert(func, name, v)

    def arg_upsert_dict(self, func: str, d: dict):
        """参数批量插入 DICT"""
        for k, v in d.items():
            self.arg_upsert(func, k, v)

    def arg_get_list(self, func: str, k: str):
        """参数获取"""
        table = self._cache.table(func) if func else self._cache
        res = table.search(self._query.__getattr__(self.KEY_NAME) == k)
        res2 = sorted(res, key=lambda item: -item[self.TIMESTAMP])
        res3 = [v[self.VAL_NAME] for v in res2]
        return res3

    def close(self):
        log.info("关闭缓存")
        self._cache.close()



from collections.abc import MutableMapping
from functools import reduce


def deepmerge(*dicts) -> dict:
    def _deepmerge(source: dict, destination: dict) -> dict:
        """Updates two dicts of dicts recursively (https://stackoverflow.com/a/24088493/8965861)."""
        for k, v in source.items():
            if k in destination:
                # this next check is the only difference!
                if all(isinstance(e, MutableMapping) for e in (v, destination[k])):
                    destination[k] = deepmerge(v, destination[k])
                # we could further check types and merge as appropriate here.
        d3 = source.copy()
        d3.update(destination)
        return d3

    return reduce(_deepmerge, tuple(dicts))



"""
create_time: 2025-03-20
author: Jacky Lee
description: 文件夹复制操作
1. 首先删除目标文件夹
2. 然后复制源文件夹
"""

# ⚠️高危操作
# TODO 测试如果 DST 存在文件不删除，直接复制过去会怎么样
import logging
import shutil
from pathlib import Path

from send2trash import send2trash

from devman.common import assert_path_exist_and_is_dir

log = logging.getLogger(__name__)


def copytree(src: Path, dst: Path, dry: bool):
    """复制文件夹

    Args:
        src (Path): 源文件夹
        dst (Path): 目标文件夹
        dry (bool): 是否干运行
    """
    log.debug(f"dry={dry}")
    if dry:
        return
    assert_path_exist_and_is_dir(src)

    if dst.exists():
        log.debug(f"目标文件夹='{dst}': 删除到回收站")
        send2trash(dst)
    try:
        shutil.copytree(src, dst, dirs_exist_ok=True)
        log.debug(f"目标文件夹='{dst}': 复制完成")
    except FileExistsError:
        log.debug(f"目标文件夹='{dst}': 已经存在同名文件，复制失败")


def copytree_struct(src: Path, dst: Path, max_depth: int = 4, file: bool = True):
    """复制文件夹结构，文件内容为 0 KB

    Args:
        src (Path): 源文件夹
        dst (Path): 目标文件夹
        max_depth (int, optional): 默认递归深度. 默认是 4.
        file (bool, optional): 是否复制文件. 默认是 True.
    """
    # 校验：防止 dst 覆盖 src
    assert src.exists(), f"{src} 源文件夹不存在，请检查路径"
    assert not dst.exists(), f"{dst} 目标文件夹已存在，请检查路径"
    # 创建空目录
    dst.mkdir(parents=True)
    log.debug(f"准备复制 {src} -> {dst}, max_depth={max_depth}, file={file}")

    # if not confirm():
    #     log.info("退出程序...")
    #     return
    queue = [(src, 0)]
    while queue:
        dir, level = queue.pop()
        if level > max_depth:  # 如果递归深度大于最大深度，跳出递归
            continue
        try:
            for p in dir.iterdir():
                # 文件夹
                parts = p.relative_to(src).parts
                dst_path = dst.joinpath(*parts)
                if p.is_dir():
                    queue.append((p, level + 1))  # 递归进入下一层
                    dst_path.mkdir()
                    continue
                # 文件
                if p.is_file() and file:
                    log.info(f"复制 {p} -> {dst_path}")
                    dst_path.touch()
        except PermissionError:
            log.info(f"无权限访问，跳过... {p}")
            continue
